unit JavaDMain;

{
  JavaD - Delphi interface to a Java class.
  JavaD builds a Pascal skeleton for using JNI with Delphi.
  Demonstrates how to call Java classes from Delphi.

  Written by Keith Wood (kbwood@iprimus.com.au).
  9 April 2003
}

interface

uses
  SysUtils, Classes, Controls, Forms, StdCtrls, ComCtrls, JNI,
{$IFDEF JNIUTILS}
  JNIUtils,
{$ENDIF}
  Dialogs;

type
  TfrmJavaD = class(TForm)
    Label1: TLabel;
    edtClassName: TEdit;
    Label2: TLabel;
    edtDirectory: TEdit;
    chkOverwrite: TCheckBox;
    btnBrowse: TButton;
    btnLoad: TButton;
    btnGenerate: TButton;
    memOutput: TMemo;
    stbStatus: TStatusBar;
    dlgOpen: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure edtChange(Sender: TObject);
    procedure btnBrowseClick(Sender: TObject);
    procedure btnLoadClick(Sender: TObject);
    procedure btnGenerateClick(Sender: TObject);
  private
    FJavaVM: TJavaVM;
    FJNIEnv: TJNIEnv;
    procedure ShowStatus(const Message: string);
  public
  end;

var
  frmJavaD: TfrmJavaD;

implementation

{$R *.dfm}

{ Initialisation }
procedure TfrmJavaD.FormCreate(Sender: TObject);
begin
  FJavaVM           := nil;
  FJNIEnv           := nil;
  edtDirectory.Text := ExtractFilePath(Application.ExeName);
end;

{ Release resources }
procedure TfrmJavaD.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FJNIEnv);
  FreeAndNil(FJavaVM);
end;

{ Can only generate when the JVM has been loaded and
  a class name and directory are present }
procedure TfrmJavaD.edtChange(Sender: TObject);
begin
  btnGenerate.Enabled := btnGenerate.Default and
    (edtClassName.Text <> '') and (edtDirectory.Text <> '');
end;

{ Find an output directory }
procedure TfrmJavaD.btnBrowseClick(Sender: TObject);
begin
  with dlgOpen do
  begin
    FileName := edtDirectory.Text + '.dpr';
    if Execute then
      edtDirectory.Text := ExtractFilePath(FileName);
  end;
end;

{ Load the JVM }
procedure TfrmJavaD.btnLoadClick(Sender: TObject);
var
  ClassPath: string;
  Errcode: Integer;
  VM_args: JavaVMInitArgs;
  Options: array [0..10] of JavaVMOption;
begin
  ShowStatus('Loading Java VM...');

  // Confirm the classpath
  ClassPath := InputBox(
    'JavaD', 'Enter the Java classpath', 'C:\Projects\DJNI\classes');

  // Set up the options for the VM
  FillChar(Options, SizeOf(Options), #0);
  Options[0].optionString := PChar('-Djava.class.path=' + ClassPath);
  VM_args.version  := JNI_VERSION_1_2;
  VM_args.options  := @Options;
  VM_args.nOptions := 1;

  // Create the wrapper for the VM
  FJavaVM := TJavaVM.Create;
  // Load the VM
  Errcode := FJavaVM.LoadVM(VM_args);
  if Errcode < 0 then
  begin
    // Loading the VM more than once will cause this error
    if Errcode = JNI_EEXIST then
      raise Exception.Create(
        'Java VM has already been loaded. Only one JVM can be loaded.')
    else
      raise Exception.Create(
        Format('Error creating JavaVM, code = %d', [Errcode]));
  end;

  // Create the Env class
  FJNIEnv := TJNIEnv.Create(FJavaVM.Env);
  ShowStatus(Format('Java VM %d.%d loaded',
    [FJNIEnv.MajorVersion, FJNIEnv.MinorVersion]));

  // Adjust buttons
  btnLoad.Enabled     := False;
  btnLoad.Default     := False;
  btnGenerate.Default := True;
  edtChange(nil);
end;

{ Generate a Pascal skeleton for JNI }
procedure TfrmJavaD.btnGenerateClick(Sender: TObject);
var
  Cls: JClass;
{$IFNDEF JNIUTILS}
  MID: JMethodID;
{$ENDIF}
  Exc: JThrowable;
  FileName: string;
  ErrMsg: string;
begin
  try
    memOutput.Lines.Clear;
    ShowStatus('Calling generator method...');

    // Find JavaD class
    Cls := FJNIEnv.FindClass('wood/keith/opentools/javad/JavaD');
    if Cls = nil then
      raise Exception.Create(
        'Can''t find class: wood.keith.opentools.javad.JavaD');

    // Run it
{$IFDEF JNIUTILS}
    FileName := JNIUtils.CallMethod(FJNIEnv, Cls,
      'generateDelphiWrapper', 'String (String, String, boolean)',
      [edtClassName.Text, edtDirectory.Text, chkOverwrite.Checked], True);
{$ELSE}
    MID := FJNIEnv.GetStaticMethodID(Cls, 'generateDelphiWrapper',
      '(Ljava/lang/String;Ljava/lang/String;Z)Ljava/lang/String;');
    if MID = nil then
      raise Exception.Create('Can''t find method: generateDelphiWrapper');

    FileName := FJNIEnv.JStringToString(FJNIEnv.CallStaticObjectMethod(
      Cls, MID, [edtClassName.Text, edtDirectory.Text, chkOverwrite.Checked]));
{$ENDIF}

    // Check for exception
    Exc := FJNIEnv.ExceptionOccurred;
    if Exc <> nil then
    begin
      // Clear the exception so we can call other methods
      FJNIEnv.ExceptionClear;
      // Find out about the exception - its class and message
{$IFDEF JNIUTILS}
      Cls    := JNIUtils.CallObjectMethod(FJNIEnv, Exc, 'getClass', 'Class()', []);
      ErrMsg := JNIUtils.CallMethod(FJNIEnv, Cls, 'getName', 'String()', []) +
        #13 + JNIUtils.CallMethod(FJNIEnv, Exc, 'getMessage', 'String()', []);
{$ELSE}
      MID := FJNIEnv.GetMethodID(
        FJNIEnv.GetObjectClass(Exc), 'getClass', '()Ljava/lang/Class;');
      if MID = nil then
        raise Exception.Create('Can''t find method: getClass');

      Cls := FJNIEnv.CallObjectMethod(Exc, MID, []);

      MID := FJNIEnv.GetMethodID(
        FJNIEnv.GetObjectClass(Cls), 'getName', '()Ljava/lang/String;');
      if MID = nil then
        raise Exception.Create('Can''t find method: getName');

      ErrMsg := FJNIEnv.JStringToString(FJNIEnv.CallObjectMethod(Cls, MID, []));

      MID := FJNIEnv.GetMethodID(
        FJNIEnv.GetObjectClass(Exc), 'getMessage', '()Ljava/lang/String;');
      if MID = nil then
        raise Exception.Create('Can''t find method: getMessage');

      ErrMsg := ErrMsg + #13 +
        FJNIEnv.JStringToString(FJNIEnv.CallObjectMethod(Exc, MID, []));
{$ENDIF}
      raise Exception.Create('A Java exception occurred'#13 + ErrMsg);
    end;

    // Load the generated file
    memOutput.Lines.LoadFromFile(FileName);

    ShowStatus('Done');
  except
    on E: Exception do
    begin
      ShowStatus('Error');
      MessageDlg('Error: ' + E.Message, mtError, [mbOK], 0);
    end;
  end;
end;

{ Display the status message }
procedure TfrmJavaD.ShowStatus(const Message: string);
begin
  stbStatus.SimpleText := Message;
end;

end.

