package wood.keith.opentools.javad;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;

/**
 * Generate Pascal skeletons for interacting with Java through JNI.
 * For use with the JEDI JNI project.
 * Also available at http://home.pacifier.com/~mmead/jni/delphi/index.html.
 *
 * @author   Keith Wood
 * @version  1.0  26 November 2002
 */
public class JavaD {

  /** Native methods that are automatically implemented for all objects. */
  private static final String[] IGNORE_METHODS = {
      "getClass", "hashCode", "notify", "notifyAll", "wait"};
  /** The mapping from Java types to Pascal types. */
  private static HashMap delphiTypes = new HashMap();
  /** The mapping from Java signature values to Java types. */
  private static HashMap javaTypes = new HashMap();
  /** The mapping from Java types to Java signature values. */
  private static HashMap sigTypes = new HashMap();

  static {
    delphiTypes.put("boolean", "JBoolean");
    delphiTypes.put("byte", "JByte");
    delphiTypes.put("char", "JChar");
    delphiTypes.put("double", "JDouble");
    delphiTypes.put("float", "JFloat");
    delphiTypes.put("int", "JInt");
    delphiTypes.put("long", "JLong");
    delphiTypes.put("short", "JShort");
    delphiTypes.put("Z", "JBoolean");
    delphiTypes.put("B", "JByte");
    delphiTypes.put("C", "JChar");
    delphiTypes.put("D", "JDouble");
    delphiTypes.put("F", "JFloat");
    delphiTypes.put("I", "JInt");
    delphiTypes.put("J", "JLong");
    delphiTypes.put("S", "JShort");
    delphiTypes.put("java.lang.String", "JString");
    javaTypes.put("Z", "boolean");
    javaTypes.put("B", "byte");
    javaTypes.put("C", "char");
    javaTypes.put("D", "double");
    javaTypes.put("F", "float");
    javaTypes.put("I", "int");
    javaTypes.put("J", "long");
    javaTypes.put("S", "short");
    sigTypes.put("boolean", "Z");
    sigTypes.put("byte",    "B");
    sigTypes.put("char",    "C");
    sigTypes.put("double",  "D");
    sigTypes.put("float",   "F");
    sigTypes.put("int",     "I");
    sigTypes.put("long",    "J");
    sigTypes.put("short",   "S");
    sigTypes.put("void",    "V");
    sigTypes.put("java.lang.String", "Ljava/lang/String;");
  }

  /**
   * Generate a Pascal skeleton file for using JNI on a given class.
   *
   * @param  className  the name of the class to interface with
   * @param  dirName    the output directory name - generated file is the
   *                    same as the class name plus .dpr
   * @param  overwrite  true to overwrite any existing file,
   *                    false to abort if existing
   * @return  the name of the generated file
   * @throws  Exception  if an error occurs
   */
  public static String generateDelphiWrapper(String className,
      String dirName, boolean overwrite) throws Exception {
    return generateDelphiWrapper(className, dirName,
      ClassLoader.getSystemClassLoader(), overwrite);
  }

  /**
   * Generate a Pascal skeleton file for using JNI on a given class.
   *
   * @param  className  the name of the class to interface with
   * @param  dirName    the output directory name - generated file is the
   *                    same as the class name plus .dpr
   * @param  loader     the class loader to find the class to parse
   * @param  overwrite  true to overwrite any existing file,
   *                    false to abort if existing
   * @return  the name of the generated file
   * @throws  Exception  if an error occurs
   */
  public static String generateDelphiWrapper(String className,
      String dirName, ClassLoader loader, boolean overwrite) throws Exception {
    StringBuffer signature = new StringBuffer();
    StringBuffer declarations = new StringBuffer();
    StringBuffer exports = new StringBuffer();
    StringBuffer vars = new StringBuffer();
    StringBuffer code = new StringBuffer();
    Class clazz = loader.loadClass(className);
    Method[] methods = clazz.getMethods();
    // Find all public native methods
    for (int index = 0; index < methods.length; index++) {
      Method method = methods[index];
      if (Modifier.isNative(method.getModifiers()) &&
          Modifier.isPublic(method.getModifiers())) {
        if (Arrays.binarySearch(IGNORE_METHODS, method.getName()) > -1) {
          continue;
        }
        // Is this method overloaded?
        boolean isOverloaded = false;
        for (int index2 = 0; index2 < methods.length; index2++) {
          if (index2 != index &&
              methods[index2].getName().equals(method.getName()) &&
              Modifier.isNative(methods[index2].getModifiers()) &&
              Modifier.isPublic(methods[index2].getModifiers())) {
            isOverloaded = true;
            break;
          }
        }

        // Include Java signature into Pascal as comment
        declarations.append("{\n  Java declaration: ").
          append(Modifier.toString(method.getModifiers())).append(" ").
          append(expandType(method.getReturnType().getName())).append(" ").
          append(method.getName()).append("(");
        signature.setLength(0);
        Class[] params = method.getParameterTypes();
        for (int index2 = 0; index2 < params.length; index2++) {
          signature.append(condenseType(params[index2].getName()));
          declarations.append(index2 > 0 ? ", " : "").
            append(expandType(params[index2].getName()));
        }
        Class[] exceptions = method.getExceptionTypes();
        declarations.append(")");
        String connector = " throws";
        for (int index2 = 0; index2 < exceptions.length; index2++) {
          declarations.append(connector).append(exceptions[index2].getName());
          connector = ", ";
        }
        declarations.append("\n  Class: ").append(className).
          append("\n  Method: ").append(method.getName()).
          append("\n  Signature: (").append(signature.toString()).
          append(")").append(condenseType(method.getReturnType().getName())).
          append("\n}\n");
        // Generate the procedure/function skeleton
        vars.setLength(0);
        code.setLength(0);
        boolean isProcedure = method.getReturnType().getName().equals("void");
        String delphiName =
          "Java_" + className.replace('.', '_') + "_" + method.getName() +
          (isOverloaded ? getOverloadedSignature(signature.toString()) : "");
        declarations.append((isProcedure ? "procedure " : "function ")).
          append(delphiName).append("(PEnv: PJNIEnv; Obj: JObject");
        for (int index2 = 0; index2 < params.length; index2++) {
          String paramType = translateType(params[index2].getName());
          declarations.append("; Param").append(index2).append(": ").
            append(paramType);
          if (paramType.indexOf("Array") > -1) {
            vars.append("  Len").append(index2).append(": Integer;\n");
            code.append("    Len").append(index2).
              append(" := JVM.GetArrayLength(Param").append(index).append(");\n");
          }
        }
        declarations.append(")");
        if (!isProcedure) {
          String returnType = translateType(method.getReturnType().getName());
          declarations.append(": ").append(returnType);
          if (returnType.equals("JString")) {
            code.append("    Result := JVM.NewStringUTF('');\n");
          }
          else if (returnType.equals("JInt") || returnType.equals("JLong") ||
              returnType.equals("JByte") || returnType.equals("JShort") ||
              returnType.equals("JDouble") || returnType.equals("JFloat")) {
            code.append("    Result := 0;\n");
          }
          else if (returnType.equals("JBoolean")) {
            code.append("    Result := JNI_FALSE;\n");
          }
          else if (returnType.equals("JChar")) {
            code.append("    Result := '';\n");
          }
          else if (returnType.equals("JObject")) {
            code.append("    Result := JVM.NewObject(Cls, MId, []);\n");
          }
          else if (returnType.equals("JObjectArray")) {
            code.append(
              "    Result := JVM.NewObjectArray(Size, Cls, Values);\n");
          }
        }
        declarations.append(";\n  {$IFDEF WIN32} stdcall; {$ENDIF}").
          append(" {$IFDEF LINUX} cdecl; {$ENDIF}\n").
          append("var\n").
          append("  JVM: TJNIEnv;\n").
          append(vars.toString()).
          append("begin\n").
          append("  JVM := TJNIEnv.Create(PEnv);\n").
          append("  try\n").
          append(code.toString()).
          append("  finally\n").
          append("    JVM.Free;\n").
          append("  end;\n").
          append("end;\n\n");
        // Add to exports list
        exports.append("  ").append(delphiName).append(",\n");
      }
    }
    if (declarations.length() == 0) {
      throw new Exception("No public native methods found in " + className);
    }
    // Actually write the Pascal file
    File file = new File(getFileName(className, dirName));
    if (file.exists() && !overwrite) {
      throw new Exception("File already exists " + file.getAbsolutePath());
    }
    PrintStream out = new PrintStream(new FileOutputStream(file));
    out.println("library " + className.replace('.', '_') + ";\n");
    out.println("{");
    out.println("  Pascal skeleton for using JNI with the " + className +
      " Java class.\n");
    out.println("  Generated " + new Date() + " by JavaD.");
    out.println("}\n");
    out.println("uses\n  JNI;\n");
    out.println(declarations.toString());
    out.println("exports");
    out.println(exports.substring(0, exports.length() - 2) + ";");
    out.println("end.");
    out.close();
    return file.getAbsolutePath();
  }

  /**
   * Encode Java type to signature value.
   *
   * @param  jType  the Java type
   * @return  the signature code
   */
  private static String condenseType(String jType) {
    String condensed = (String)sigTypes.get(jType);
    return (condensed == null ? jType.replace('.', '/') : condensed);
  }

  /**
   * Decode Java type signature to full text.
   *
   * @param  jType  the encoded type signature
   * @return  the full text version
   */
  private static String expandType(String jType) {
    int arrayDims = 0;
    while (jType.charAt(0) == '[') {
      arrayDims++;
      jType = jType.substring(1);
    }
    String expanded = (String)javaTypes.get(jType);
    if (expanded == null) {
      expanded = (jType.charAt(0) == 'L' ?
        jType.substring(1, jType.length() - 1) : jType);
    }
    while (arrayDims-- > 0) {
      expanded = expanded + "[]";
    }
    return expanded;
  }

  /**
   * Mangle the method parameter signature to distinguish overloaded methods.
   *
   * @param  signature  the parameter signature
   * @return  the mangled signature for overloading
   */
  private static String getOverloadedSignature(String signature) {
    StringBuffer buffer = new StringBuffer(signature);
    for (int index = buffer.length() - 1; index >= 0; index --) {
      if (buffer.charAt(index) == '/') {
        buffer.replace(index, index + 1, "_");
      }
      else if (buffer.charAt(index) == '_') {
        buffer.replace(index, index + 1, "_1");
      }
      else if (buffer.charAt(index) == ';') {
        buffer.replace(index, index + 1, "_2");
      }
      else if (buffer.charAt(index) == '[') {
        buffer.replace(index, index + 1, "_3");
      }
    }
    return buffer.insert(0, "__").toString();
  }

  /**
   * Convert Java type to Pascal type.
   *
   * @param  jType  the expanded Java type
   * @return  the Pascal type
   */
  private static String translateType(String jType) {
    boolean isArray = jType.startsWith("[");
    if (isArray) {
      jType = jType.substring(1);
    }
    String dType = (String)delphiTypes.get(jType);
    if (dType == null) {
      dType = "JObject";
    }
    return dType + (isArray ? "Array" : "");
  }

  /**
   * Construct the new file name from the class and directory provided.
   *
   * @param  className  the name of the class to interface with
   * @param  dirName    the output directory name
   * @return  the generated file name - same as the class name plus .dpr
   */
  static String getFileName(String className, String dirName) {
    dirName = dirName.replace('\\', '/');
    return dirName + (dirName.charAt(dirName.length() - 1) != '/' ? "/" : "") +
      className.replace('.', '_') + ".dpr";
  }

  /**
   * Run it from the command line.
   */
  public static void main(String[] args) {
    boolean overwrite = (args.length == 3 && args[0].equals("-o"));
    if (!(args.length == 2 || overwrite)) {
      System.out.println("Usage: JavaD [-o] <class> <output directory>");
      return;
    }
    try {
      int base = (overwrite ? 1 : 0);
      System.out.println("Generated " +
        generateDelphiWrapper(args[base], args[base + 1], overwrite));
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
}